from draughts.core.variant import Board, Move, WHITE, BLACK

__all__ = ["Board", "Move", "WHITE", "BLACK"]

__author__ = "Ioannis Pantidis"
__copyright__ = "2021-2024, " + __author__
__version__ = "0.6.6"
